/*
 Stephan Kottler
 kottlers@informatik.uni-tuebingen.de
 *
 */

#ifndef BITVECTOR_HPP_
#define BITVECTOR_HPP_
#define bitvector_hpp_version  "$Revision: 1.2 $"

#include <stdint.h>
#include <stdlib.h>
#include <string.h>
#include "Assertion.hpp"


template<class KeyType>
struct DfltKeyTypeConverter{
	uint32_t toInt(const KeyType& k)const {return k.idx();}
	KeyType fromInt(const uint32_t& i)const {return KeyType(i);}
};

//specialisation
template<>
struct DfltKeyTypeConverter<uint32_t>{
	uint32_t toInt(const uint32_t& k)const {return k;}
	uint32_t fromInt(const uint32_t& i)const {return i;}
};


//# if __WORDSIZE == 64
#define BITVEC_OFFSET_MASK	63
#define BITVEC_BIT_SHIFT 	6
#define BITVEC_DATATYPE 	uint64_t
/*#else
#define BITVEC_OFFSET_MASK 	31
#define BITVEC_BIT_SHIFT 	5
#define BITVEC_DATATYPE 	uint32_t
#endif*/

/**
 * T is the type of elements for which flags can be set.
 * It has to implement the function T.idx() to get an integer representing an element.
 */
#define BVIDX asrt(idx<sze);asrt(ar_idx(idx)<array_sz);
#define BVE asrt(conv.toInt(e)<sze);asrt(ar_idx(conv.toInt(e))<array_sz);
template<class E, class EleToIdx = DfltKeyTypeConverter<E> >
class BitVector{
protected:
	EleToIdx conv;

	/**Array to store a flag-value for each element.*/
	BITVEC_DATATYPE*   bits;

	/**Number of elements that are hold.*/
	uint32_t	sze;

	uint32_t	array_sz;

	//construction, assignment, destruction

	/**No copying allowed*/
private:
	BitVector(const BitVector& flg)					{asrt(false);};
protected:
	const static BITVEC_DATATYPE zero	= 0;
	const static BITVEC_DATATYPE one	= 1;


	inline static BITVEC_DATATYPE	offset(BITVEC_DATATYPE index)
		{return index & BITVEC_OFFSET_MASK ;}

	inline static BITVEC_DATATYPE	off_mask(BITVEC_DATATYPE index)
		{return one << (index & BITVEC_OFFSET_MASK) ;}

	inline static BITVEC_DATATYPE	ar_idx(BITVEC_DATATYPE index)
		{return index >> BITVEC_BIT_SHIFT ;}


public:
	uint64_t getMemTotal() const					{return array_sz * 8;}

	inline BitVector(const BitVector& flg,const bool& deep);
	inline BitVector(): bits(NULL),sze(0),array_sz(0){}
	//inline BitVector(uint32_t nbr_elements);
	inline void hardClear();
	inline ~BitVector(){hardClear();}

	inline const BitVector& operator=(const BitVector& flg);

	inline void ensureSize(const uint32_t nbr_elements,const bool copy);
	inline bool isSetAt(const uint32_t& idx)	const 	{BVIDX return bits[ar_idx(idx)] & off_mask(idx);}
	inline bool isSet(const E& e)	const 				{return isSetAt(conv.toInt(e));}
	inline void setFlagAt(const uint32_t& idx)			{BVIDX bits[ar_idx(idx)] |= off_mask(idx);}
	inline void setFlag(const E& e)						{setFlagAt(conv.toInt(e));}
	inline void flipAt(const uint32_t& idx)				{BVIDX bits[ar_idx(idx)] ^= off_mask(idx);}
	inline void flip(const E& e)						{flipAt(conv.toInt(e));}
	inline void clearAt(const uint32_t& idx)			{BVIDX bits[ar_idx(idx)] &= ~off_mask(idx);}
	inline void clear(const E& e)						{BVE clearAt(conv.toInt(e));}
	inline void carelessClear(const E& e)	{BVE bits[ar_idx(conv.toInt(e))] = zero;}
	inline void clearAll()					{for(uint32_t i(array_sz);i--;){bits[i] = zero; }};
	inline uint32_t size() const			{return sze;}
	inline bool setFlagNew(const E& e)		{BVE BITVEC_DATATYPE& f	= bits[ar_idx(conv.toInt(e))];
											 BVE BITVEC_DATATYPE m 	= off_mask(conv.toInt(e));
											 bool was=f&m; f|=m; asrt(f&m); return !was;}
	inline bool allCleared();
	//inline bool checkAndSet(const E& e);
	inline bool checkAndClear(const E& e);

};

template<class E, class Idx=DfltKeyTypeConverter<E> >
class TwoBitsVec:private BitVector<E,Idx>{
	static const uint64_t three = 3;
	typedef BitVector<E,Idx> sup;
public:
	uint64_t getMemTotal() const				{return sup::getMemTotal();}
	TwoBitsVec():sup(){}
	TwoBitsVec(const TwoBitsVec& flg,const bool& deep):sup(flg,deep){}

	void ensureSize(const uint32_t els,const bool copy){sup::ensureSize(els<<1,copy);}
	uint32_t getSize()					const	{return sup::sze>>1;}
	bool firstSet(const E& e)	const 			{return sup::isSetAt((sup::conv.toInt(e)<<1)+1);}
	bool secSet(const E& e)	const 				{return sup::isSetAt(sup::conv.toInt(e)<<1);}
	bool anySet(const E& e)			const		{uint32_t idx = sup::conv.toInt(e)<<1;
												return ((sup::bits[sup::ar_idx(idx)]>>sup::offset(idx))&three);}
	bool hasVal(const E& e, uint64_t v) const{asrt(v<4); uint32_t idx = sup::conv.toInt(e) <<1;
												return ((sup::bits[sup::ar_idx(idx)]>>sup::offset(idx))&three)==v;}

	void setFirst(const E& e)					{sup::setFlagAt((sup::conv.toInt(e)<<1)+1);}
	void setSec(const E& e)						{sup::setFlagAt(sup::conv.toInt(e)<<1);}

	void flipFirst(const E& e)					{sup::flipAt((sup::conv.toInt(e)<<1)+1);}
	void flipSec(const E& e)					{sup::flipAt(sup::conv.toInt(e)<<1);}

	void clearFirst(const E& e)					{sup::clearAt((sup::conv.toInt(e)<<1)+1);}
	void clearSec(const E& e)					{sup::clearAt(sup::conv.toInt(e)<<1);}
	void clearBoth(const E& e)					{uint32_t idx= sup::conv.toInt(e)<<1; sup::bits[sup::ar_idx(idx)] &= ~(three<< sup::offset(idx));}

	void setVal(const E& e, uint64_t v) 		{asrt(v<4);
		uint32_t idx=sup::conv.toInt(e)<<1; uint64_t& b=sup::bits[sup::ar_idx(idx)]; uint32_t of= sup::offset(idx);
		b &= ~(three<<of); b|= (v<<of);}

	uint64_t setGetVal(const E& e, uint64_t v) 		{asrt(v<4);
		uint32_t idx	= sup::conv.toInt(e)<<1;
		uint64_t& b	= sup::bits[sup::ar_idx(idx)];
		uint32_t of	= sup::offset(idx);
		uint64_t was	= (b >> of) & three;
		b 			&= ~(three<<of);
		b			|= (v<<of);
		return was;
	}

	void setValIf(const E& e,uint64_t exp, uint64_t v) {asrt(v<4); asrt(exp<4);
		uint32_t idx=sup::conv.toInt(e)<<1; uint64_t& b=sup::bits[sup::ar_idx(idx)]; uint32_t of= sup::offset(idx);
		uint64_t msk= three<<of;
		if((b & msk) == (exp<<of)){b &= ~(msk); b|= (v<<of);}}

	void addVals(const E& e, const bool first, const bool sec){
		uint32_t idx=sup::conv.toInt(e)<<1; uint64_t& b=sup::bits[sup::ar_idx(idx)]; uint32_t of= sup::offset(idx);
		uint64_t v = (((uint64_t)first) << 1) | (uint64_t)sec;
		b|= (v<<of);}

	void hardClear()							{sup::hardClear();}
	bool allCleared()							{return sup::allCleared();}
	void clearAll(){sup::clearAll();}
};

template<class E, class EleToIdx>
BitVector<E,EleToIdx>::BitVector(const BitVector<E,EleToIdx>& flg,const bool& deep)
	:bits(flg.bits),sze(flg.sze),array_sz(flg.array_sz){

	if(deep && bits!=NULL){
		bits = (BITVEC_DATATYPE*)malloc( array_sz * sizeof(BITVEC_DATATYPE));
		asrt(bits != NULL);
		memcpy(bits,flg.bits,array_sz* sizeof(BITVEC_DATATYPE));
	}
}


//Inline Implementation

template<class E, class EleToIdx>
inline void BitVector<E,EleToIdx>::hardClear(){
	if(bits != NULL){
		free(bits);
		bits		= NULL;
		sze			= 0;
		array_sz	= 0;
	}
}

template<class E, class EleToIdx>
const BitVector<E,EleToIdx>& BitVector<E,EleToIdx>::operator=(const BitVector<E,EleToIdx>& flg){
	if(&flg == this ) return *this;
	if(bits != NULL){free(bits);}
	sze 		= flg.sze;
	array_sz	= flg.array_sz;
	bits 		= (BITVEC_DATATYPE*)malloc( array_sz * sizeof(BITVEC_DATATYPE));

	memcpy(bits,flg.bits,array_sz* sizeof(BITVEC_DATATYPE));
	return *this;
}

template<class E, class EleToIdx>
inline void BitVector<E,EleToIdx>::ensureSize(const uint32_t new_sz,const bool copy){
	if(new_sz <= sze) return;
	asrt(new_sz > 0);

	uint32_t new_arsz 	= ar_idx(new_sz)+(offset(new_sz) > 0);
	if(new_arsz <= array_sz) {sze = new_sz;return;}


	if(copy && bits != NULL){
		asrt(new_arsz >= array_sz);

		BITVEC_DATATYPE* new_bits 	= (BITVEC_DATATYPE*)malloc( new_arsz * sizeof(BITVEC_DATATYPE));
		asrt(new_bits != NULL);
		memcpy(new_bits,bits,array_sz* sizeof(BITVEC_DATATYPE));
		free(bits);
		memset(new_bits + array_sz, 0, (new_arsz-array_sz) * sizeof(BITVEC_DATATYPE) );

		sze 		= new_sz;
		array_sz 	= new_arsz;
		bits		= new_bits;

	}else{

		if(bits != NULL){free(bits);}

		sze 		= new_sz;
		array_sz 	= new_arsz;

		bits   = (BITVEC_DATATYPE*)malloc( array_sz * sizeof(BITVEC_DATATYPE));
		memset ( bits, 0, array_sz * sizeof(BITVEC_DATATYPE) );
	}
}

/*
template<class E, class EleToIdx>
bool BitVector<E,EleToIdx>::checkAndSet(const E& e){
	BVE
	const uint64_t msk	= off_mask(conv.toInt(e));
	uint64_t& fld  		= bits[ar_idx(conv.toInt(e))];
	bool before 		= fld & msk;
	fld 				|= msk;
	return before;
}
*/

template<class E, class EleToIdx>
bool BitVector<E,EleToIdx>::checkAndClear(const E& e){
	BVE
	bool before = isSet(e);
	clear(e);
	return before;

}

template<class E, class EleToIdx>
bool BitVector<E,EleToIdx>::allCleared(){
	for(uint32_t i(array_sz);i--;){
		if(bits[i] != 0) {return false;}
	}
	return true;
}












#endif /*BITVECTOR_HPP_*/
